/*

    fullcalendar and other basic adjustments for seantis.reservation

*/

/* the approve field is now called approve_manually, since the old one has to linger around for migration
and there's no easy way to hide it, this hack helps: */
#formfield-form-widgets-approve {
    display: none;
}

/* without this the events in the month view are not clickable */
.fc-view-month .fc-day-content {
    visibility: hidden;
}


/* limited list */
.limitedList {
    overflow: auto;
    max-height: 150px;
    max-width: 400px;
    position: relative;
}

.your-reservations .limitedList,
.your-reservations-view .limitedList {
    max-width: 100%;
}

/* controlpanel settings page */
#formfield-form-widgets-send_email_to_managers .option {
    display: block;
}

/* overview calendar */
#seantis-overview-calendar .fc-widget-header.fc-mon,
#seantis-overview-calendar .fc-widget-header.fc-tue,
#seantis-overview-calendar .fc-widget-header.fc-wed,
#seantis-overview-calendar .fc-widget-header.fc-thu,
#seantis-overview-calendar .fc-widget-header.fc-fri {
    background-color: #efefef;
    border: none;
}

#seantis-overview-calendar .fc-widget-header.fc-sat,
#seantis-overview-calendar .fc-widget-header.fc-sun {
    background-color: #ddd !important;
    border: none;
}


.groupSelection {
    -webkit-box-shadow: 0 0 0 1px #0e6700 !important;
    -moz-box-shadow: 0 0 0 1px #0e6700 !important;
    box-shadow: 0 0 0 1px #0e6700 !important;
    background-color: white !important;
    position: relative;
}

.groupclickable:hover {
    cursor: pointer;
}

.event-unavailable.groupSelection {
    -webkit-box-shadow: 0 0 0 1px #ae1e00 !important;
    -moz-box-shadow: 0 0 0 1px #ae1e00 !important;
    box-shadow: 0 0 0 1px #ae1e00 !important;
}

.event-partly-available.groupSelection {
    -webkit-box-shadow: 0 0 0 1px #dd7c07 !important;
    -moz-box-shadow: 0 0 0 1px #dd7c07 !important;
    box-shadow: 0 0 0 1px #dd7c07 !important;
}

#seantis-overview-calendar .groupSelection {
    -webkit-box-shadow: none !important;
    -moz-box-shadow: none !important;
    box-shadow: none !important;
}

a.fc-event,
a.fc-event:hover {
    text-decoration: none;
    color: white !important;
    border-radius: 0 !important;
}

.fc-event-title {
    font-size: 10px;
    line-height: 14px;
}

/* availability colors */
body .event-available {
    background-color: #0e6700 !important;
    border-color: #0e6700 !important;
}
body .event-available .fc-event-time {
    background-color: #0d670b;
}
body .event-partly-available {
    background-color: #dd7c07 !important;
    border-color: #dd7c07 !important;
}
body .event-partly-available .fc-event-time {
    background-color: #dd7c18;
}

body .event-unavailable {
    background-color: #ae1e00 !important;
    border-color: #ae1e00 !important;
}
body .event-unavailable .fc-event-time {
    background-color: #ae1e09;
}

/* corner icon */
.fc-event-time {
    line-height: 19px !important;
}

.fc-event-time span {
    float: left;
}

.approve-manually .corner-icon {
    float: right;
    width: 13px;
    height: 13px;
    margin: 3px 2px 0 0;
}

.approve-manually .corner-icon,
.waitinglist-icon {
    background: url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAA0AAAANCAYAAABy6+R8AAAAyklEQVQoz2NgQAU2QLxcVFT0iZaW1k8QDeJDxTEACxBP9fHx+X/q1Kn/yADEB4mD5IGYFVnT7Pr6+v/4QF1dHUjjbJgGa5BJ+MC5c+f+h4SEwGy0BmlaBnNSaGjo/wsXLmBoUFVVBdMgdWA/Aj37AqYApAEYAHCNIIUqKipgGgaA6p8ygBQhA5jGlStXYmgAAR0dnZ8oNiFrlJGRwdAAtwnkRvRgxgXgfgJFHKHQgwFo6MEjegqR8TQFOXJZyUkR6GnvKTTtPcWW9gD/SpZeK7AzIwAAAABJRU5ErkJggg==) no-repeat;
}

/* calendar colors */
#content .fc-content .fc-today {
    background-color: #fff0b2;
}

.fc-content td, .fc-content th {
    border-color: #ddd;
}

.fc-event-bg {
    opacity: 0.2 !important;
}

.calendar-occupied {
    background-image: url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABoAAAAaCAIAAAAmKNuZAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAgUlEQVRIia3Uyw3AIAwDULcdhbmYLoOmFyQECuTrs/Xkkx9mhiFEZKm9hVbvXefsFtR1LkvhvNaNC1hHLmbJXNgSuIy1c0lr4fLW5EqswVVZAL7WWpUF4wUYLSLSObsFdZ3LUjivdeMC1pGLWTIXtgQuY+1c0lq4vDW5EmtwVRaAHw+qTptplxYcAAAAAElFTkSuQmCC);
    background-repeat: repeat;
    border: solid #a2a2a2 1px;
}

.loading {
    background-image: url(/++resource++seantis.reservation.images/loading.gif);
    background-repeat: no-repeat;
    background-position: center 12px;
    min-height: 30px;
}

.event-unavailable.groupSelection {
    -webkit-box-shadow: 0 0 4px 1px #a1291e;
    -moz-box-shadow: 0 0 4px 1px #a1291e;
    box-shadow: 0 0 4px 1px #a1291e;
}

.event-available.groupSelection {
    -webkit-box-shadow: 0 0 4px 1px #379a00;
    -moz-box-shadow: 0 0 4px 1px #379a00;
    box-shadow: 0 0 4px 1px #379a00;
}

.event-partly-available.groupSelection {
    -webkit-box-shadow: 0 0 4px 1px #e99623;
    -moz-box-shadow: 0 0 4px 1px #e99623;
    box-shadow: 0 0 4px 1px #e99623;
}

/* inline load close button */
#inlineClose {
    background: url(/++resource++seantis.reservation.images/close.png) no-repeat;
    cursor: pointer;
    position: relative;
    height: 36px;
    width: 30px;
    float: right;
    top: 3px;
    right: 3px;
}

/* manage reservations splitscreen */

.splitscreen {
    width: 49%;
    float: left;
}

.splitleft {
    padding-right: 15px;
}

/* hidden waitinglist in manage reservations */

.hide-waitinglist .splitleft {
    width: 75%;
}

.hide-waitinglist .splitright {
    display: none;
}

/* to overlap the overlay */
#calroot {
    z-index: 10000;
}

/* ensures that the plone menu is displayed above the calendar */
.fc-view > div {
    z-index: 0 !important;
}

.fc-sat, .fc-sun {
    background-color: #eee;
}

/* calendar legend */
.legend {
    padding: 5px 0 5px 0 !important;
    background-color: #eee;
    width: 70%;
    margin-left: auto;
    margin-right: auto;
    text-align: center;
}

#content .color .occupied {
    border-style: solid;
    border-width: 8px;
    width: 0;
    height: 0;
}

#content .color .event-available {
    background-color: #5bad5b !important;
}

#content .color .event-partly-available {
    background-color: #fdad5b !important;
}

#content .color .event-unavailable {
    background-color: #d65b5b !important;
}

#content .color .event-available.occupied {
    border-left-color: #5bad5b !important;
    border-top-color: #5bad5b !important;
}

#content .color .event-partly-available.occupied {
    border-left-color: #fdad5b !important;
    border-top-color: #fdad5b !important;
}

#content .color .event-unavailable.occupied {
    border-left-color: #d65b5b !important;
    border-top-color: #d65b5b !important;
}

.legend .mini-note {
    color: #777;
}

.legend div {
    display: inline;
}

.legend > div {
    margin-right: 8px;
}

.legend div.color div {
    width: 16px;
    height: 16px;
    display: inline-block;
    position: relative;
    top: 3px;
}

.legend div.icon {
    padding: 0;
    margin: 0;
}

.legend div.icon div {
    width: 13px;
    height: 13px;
    display: inline-block;
    padding: 0;
    position: relative;
    top: 2px;
}

@media all and (max-width: 900px) {
    .singlecalendar .corner-icon {
        background-image: none !important;
    }

    .singlecalendar .legend .icon {
        display: none !important;
    }

    .singlecalendar .legend > div {
        display: block;
    }
    .singlecalendar .legend {
        text-align: left;
        width: 100%;
    }
}

@media all and (max-width: 2000px) {
    .multicalendar .corner-icon {
        background-image: none !important;
    }

    .multicalendar .legend .icon {
        display: none !important;
    }
}

/* make the daily view calendar smaller than the rest */
.seantis-reservation-view .fc-view-agendaDay {
    width: 70%;
    margin-left: auto;
    margin-right: auto;
}


/* hide the above-content viewlet in the thank you page */
.template-thank-you-for-reserving .documentDescription {
    display: none;
}

/* timespans */

.timespan-dates {
    float: left;
    margin-right: 1em;
}

.timespan-dates.highlight {
    color: #00a5b8;
}

/* search */
.result-time {
    cursor: pointer;
}

.result-hint {
    width: 16px;
    height: 16px;
}

.is-extra-result {
    background: url('data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAA6UlEQVQ4T2NkwAQ6QKFkIHYFYnGo9AsgvRuI5wDxNWQtjEgcNiC7D4gzgJgZi8Egob9APBWIS4D4N0gAZgBI8xaorTj0ogjvBPL8gPgXzICZQE4aFp0w+f9Y5KYBxbJBCnSB+AIQM5FoAMg7eiADJgNxDg5343MBSMtEkAJQqGqSacAVkAGfgZiHTAM+gwz4AsTclBhAiReuglwASjyFxEQ+FjU9IAO0gfgiEGNLffhi4Q8oCcAUTABy8klMBz1A9aUwA1iBHFBSdiPSKzuA6nyB+A/VMhOyxbDsDHKNElTiLpDeBcQY2RkAFSEyCXfgzpMAAAAASUVORK5CYII=');
    background-repeat: no-repeat;
    position: relative;
    top: 1px;
    cursor: help;
}

body.show-tooltip #miniTip_c {
    color: #444;
    padding: 1em;
}