/*

    izug basetheme report style

*/

/* controlbox */
.controlbox {
    background-color: #fff;
    border: 1px solid #ccc;
    box-shadow: none;
    border-color: #bbb;
    margin: 10px 0 15px;
}

.controlbox .nav-block {
    float: left;
    margin: 10px 20px 15px;
}

.controlbox h2 {
    font-size: 1.3em;
    font-weight: normal;
    border-bottom: 1px solid #aaa;
    min-width: 120px;
    border-color: #454545;
    margin: 10px 0;
}

.controlbox h2:first-child {
    margin-top: 4px;
}

.controlbox label {
    font-weight: normal;
    display: block;
    cursor: pointer;
}

.controlbox label span {
    color: #333;
    padding: 3px;
}

.controlbox label:hover span {
    color:#205C90;
}

.controlbox input {
    margin-right: 5px;
}

/* columns */

.monthly-report .leftcolumn {
    width: 25%;
    float: left;
}

.monthly-report .rightcolumn {
    width: 75%;
    float: right;
}

/* monthly report day group */
.monthly-report .day {
    margin: 23px 0 30px;
}

.monthly-report .day:first-child {
    margin: 5px 0 30px;
}

.monthly-report .day h2 {
    border: 3px solid #5d768f;
    border-width: 0 0 3px;
    font-size: 16px;
    padding: 2px 20px;
    padding-left: 0;
    color: black;
}

.monthly-report .day h3 {
    font-size: 14px;
    margin-bottom: 15px;
}

/* monthly report resource group */

.monthly-report .resource {
    margin: 5px 0 0 20px;
    padding: 0;
}

.monthly-report .resource h3 {
    font-size: 1.4em;
    margin: 20px 0 5px;
}

/* monthly report reservation type group */
.monthly-report .reservation-type {
    margin-bottom: 10px;
}

.monthly-report .reservation-type .reservation-header {
    border: 1px solid #ccc;
}

.monthly-report .approved.reservation-type .reservation-header {
    height: 100%;
    background-color: #4582bf;
}

.monthly-report .pending.reservation-type .reservation-header {
    height: 100%;
    background-color: #888;
}

.monthly-report .reservation-type .reservation-header h4 {
    font-size: 1.1em;
    font-weight: normal;
    color: white;
    padding: 2px 0 0 10px;
}

/* single reservation */

.monthly-report .reservation {
    border: 1px solid #ccc;
    border-width: 0 1px 1px;
    margin-bottom: 0;
}

.monthly-report .reservation.odd,
.day .reservation.odd {
    background-color: #efefef;
}


.monthly-report .reservation:hover {
    cursor: pointer;
    background-color: #dfdfdf!important;
    background-image: 0;
}

/* reservation header */
.reservation-header {
    height: 24px!important;
    border: none!important;
    border-bottom: 1px solid #dedede!important;
    font-size: 12px;
    background-color: white!important;
    border-color: #99C5E3!important;
}

.reservation-header h4 {
    height: 19px!important;
    color: #00538c!important;
    font-weight: bold!important;
}

.pending .reservation-header h4 {
    color: #454545!important;
}

/* reservation summary */

.monthly-report .reservation-type .reservation-summary {
    padding-bottom: 5px;   
}

.monthly-report .reservation-type .reservation-summary .leftcolumn a {
    margin-left: 10px;
    line-height: 25px;
}

.monthly-report .reservation-type .reservation-summary .leftcolumn div {
    margin-left: 10px;
}

/* reservation detail */

.monthly-report .reservation .reservation-detail {
     display:none;   
}

.monthly-report .reservation.show-details .reservation-detail {
    display: block;
    margin: 0;
}

.monthly-report .reservation-detail {
    padding-bottom: 10px;
    margin-top: -10px;
    position: relative;
    min-height: 27px;
}

/* reservation detail data*/

.monthly-report .reservation-detail .reservation-data {
    float: none;
    clear: both;
    width: 100%;
    margin-left: 10px;
}

.monthly-report .reservation-detail .reservation-data .block-title {
    font-weight: bold;
    margin: 10px 0 1px;
}

.monthly-report .reservation-detail .reservation-data .block-title:first-child {
    padding-top: 10px;
}

/* reservation urls */

.monthly-report .reservation .rightcolumn .reservation-urls {
    position: absolute;
    bottom: 0;
    right: 0;
    margin: 0 7px 5px;
}

.monthly-report .reservation .reservation-urls a {
    margin-left: 5px;
}

/* occupied timetable bar */

.monthly-report .day .timetable div.occupied {
    border-radius: 0;
    line-height: 16px;
}

.approved .timespan div.occupied {
    background-color: #00538c!important;
    border: none!important;
}

.pending .timespan div.occupied {
    background-color: #454545!important;
    border: none!important;
}

/* timetable */

.monthly-report table {
    font-family: Verdana, Arial, Helvetica, sans-serif!important;
}

.reservation table td {
    padding-left: 4px!important;
}

.reservation-header table {
    margin-left: 1px!important;
}

.monthly-report .day .reservation-header table.timetable th {
    font-weight: normal;
    color: black;
    background-color: white;
    height: 17px;
    border-color: #006fba!important;
    border-width: 0 0 0 1px;
}

.monthly-report .day table.timetable {
    border-color: #99C5E3!important;
    border-width: 0 0 0 1px;
}

.monthly-report .day .reservation-summary table.timetable td {
    border-bottom: 1px solid #ccc;
    border-right: none!important;
    border-left: none;
    border-color: #99C5E3!important;
}

.monthly-report .day .reservation.show-details table.timetable td {
    border-bottom: 1px solid #ccc;
    border-color: #99C5E3!important;
}

.monthly-report .day .timetable .empty_row td {
    visibility: hidden!important;
    height: 0!important;
    border-color: #99C5E3!important;
    border-width: 0 1px!important;
}

.monthly-report .day .approved .timetable div.occupied {
    background-color: #4582bf;
    border-color: #3b70a4;
}

.monthly-report .day .pending .timetable div.occupied {
    background-color: #888;
    border-color: #777;
}

.reservation,#content .monthly-report table.timetable,#content .monthly-report table.timetable tr td {
    border-color: #dedede!important;
}